def Estima(result, model_structure, modelling_settings, estimation_settings, round, framework_settings, data, run_solver) -> None: ...
def parameter_ranking(Z) -> None: ...
def parameter_selection(n_parameters, ranking_known, model_structure, modelling_settings, estimation_settings, solvera, round, framework_settings, data, run_solver) -> None: ...
