def fun_globalizer(func_name) -> None: ...
def run_mbdoe_md(design_settings, model_structure, modelling_settings, core_num, framework_settings, round) -> None: ...
def run_mbdoe_pp(design_settings, model_structure, modelling_settings, core_num, framework_settings, round) -> None: ...
def write_excel(df_combined, experiment_number, excel_path) -> None: ...
def read_excel(file_path) -> None: ...
def save_rounds(j, ranking, k_optimal_value, rCC_values, J_k_values, result, theta_parameters, design_type, round_data, modelling_settings, scaled_params, estimation_settings, solver_parameters, framework_settings, obs, case, data_storage, model_structure) -> None: ...
def save_to_jac(results, filename) -> None: ...
def load_from_jac(filename) -> None: ...
def data_appender(df_combined, experiment_number, data_storage) -> None: ...
def add_norm_par(modelling_settings) -> None: ...
