from _typeshed import Incomplete as Incomplete

def Sensa(GSA_settings, modelling_settings, model_structure, framework_settings) -> None: ...
def sample(settings, N, calc_second_order: bool = True, scramble: bool = True, skip_values: int = 0, seed: Incomplete | None = None): ...
