def PP(design_settings, model_structure, modelling_settings, core_number, framework_settings, round) -> None: ...
def pp_objective_function(x, nd, tv_iphi_vars, tv_iphi_max, ti_iphi_vars, ti_iphi_max, tv_ophi_vars, ti_ophi_vars, active_solvers, theta_parameters, tv_iphi_cvp, tf, eps, mutation, V_matrix, design_criteria, index_dict, model_structure, modelling_settings) -> None: ...
